import 'dart:io';

import 'package:copy_with_extension/copy_with_extension.dart';
import 'package:healthrian_ble_interface/healthrian_ble_interface.dart';
import 'package:healthrian_ble_interface/src/extension/extension.dart';
import 'package:permission_handler/permission_handler.dart';
import 'package:rxdart/rxdart.dart';

part 'module_ble.g.dart';

@CopyWith()
class ModuleBle {
  ModuleBle({
    required UUID this.uuid,
    this.commands = const Commands(),
    this.scanFilter = Builtin.scanFilterWearEcg,
    this.setDevice,
    this.isSimulator = false,
  });

  final Commands commands;

  BluetoothDevice? device;

  /// TODO: simulator
  final bool isSimulator;
  bool isSimulatorConnected = false;

  final void Function(BluetoothDevice device)? setDevice;

  bool Function(BluetoothDevice device) scanFilter;

  // bool Function(ScanResult scanResult) get _scanResultFilter =>
  //     (scanResult) => scanFilter(scanResult.device);
  //
  // int _sortByRssi(a, b) => b.rssi - a.rssi;

  /// SET * connectConfirm, setSerialNumber, setMode, setFirmUpdate
  Future<void> connectConfirm() async {
    final control = points[Characteristics.CONTROL_POINT];
    await control.writePacket(commands.connectConfirm);
  }

  Future<void> setSerialNumber(List<int> data) async {
    final control = points[Characteristics.CONTROL_POINT];
    await control.writePacket(commands.setSerialNumber(data));
  }

  Future<void> setMode(List<int> data) async {
    final control = points[Characteristics.CONTROL_POINT];
    await control.writePacket(commands.setMode(data));
  }

  Future<void> setFirmUpdate(List<int> data) async {
    final control = points[Characteristics.CONTROL_POINT];
    await control.writePacket(commands.setFirmUpdate(data));
  }

  /// GET * PRODUCT_MANUFACTURER, FIRMWARE_VERSION, SERIAL_NUMBER,
  Future<List<int>?> getManufacturer() async =>
      await points[Characteristics.PRODUCT_MANUFACTURER]?.read();

  Future<List<int>?> getFirmwareVersion() async =>
      await points[Characteristics.FIRMWARE_VERSION]?.read();

  Future<List<int>?> getSerialNumber() async =>
      await points[Characteristics.SERIAL_NUMBER]?.read();

  final streams = StreamImpl();

  final UUID uuid;
  final points = <Enum, BluetoothCharacteristic>{};

  Future<void> _requestPermission() async {
    if (Platform.isWindows) return;
    final permissions = [
      Permission.bluetoothScan,
      Permission.bluetooth,
      Permission.bluetoothConnect,
      Permission.bluetoothAdvertise,
    ];
    await permissions.request();
    'Permissions are request: $permissions'.log;
  }

  Future<bool> connectAndRegister(
      {String targetDevice = '', bool isStrict = false}) async {
    final isConnected =
        await connectDevice(targetDevice: targetDevice, isStrict: isStrict);

    if (isConnected) await registerCharacteristics();

    return isConnected;
  }

  Future<List<BluetoothDevice>> findConnectedDevices({
    bool isFiltered = true,
  }) async {
    await _requestPermission();

    final connected = await FlutterBluePlus.systemDevices;
    if (isFiltered) {
      return connected.where(scanFilter).toList();
    }
    return connected;
  }

  Future<List<ScanResult>> findScannedDevices({bool isFiltered = true}) async {
    await _requestPermission();

    final scanned = <ScanResult>[];
    Set<DeviceIdentifier> seen = {};

    final subscription = FlutterBluePlus.scanResults.listen((results) {
      for (final r in results) {
        if (!seen.contains(r.device.remoteId)) {
          seen.add(r.device.remoteId);
          scanned.add(r);
        }
      }
    });

    await FlutterBluePlus.startScan();

    await Future.delayed(const Duration(seconds: 2));

    await FlutterBluePlus.stopScan();
    subscription.cancel();

    if (isFiltered) {
      return scanned.where((r) => scanFilter(r.device)).toList();
    }
    return scanned;
  }

  Future<bool> connectDevice(
      {String targetDevice = '', bool isStrict = false}) async {
    if (isSimulator) {
      isSimulatorConnected = true;
      return true;
    }

    /// 0. initialize
    device = null;
    'Device is set to $device'.log;

    /// 1. devices which are connected
    final connected = await findConnectedDevices();

    device = connected
        .where((device) => device.remoteId.str == targetDevice)
        .firstOrNull;
    if (device != null) return true;

    if (!isStrict && device == null) {
      device = connected.firstOrNull;
    }
    if (device != null) return true;

    /// 2. perform scanning process
    final scanned = await findScannedDevices();

    device = scanned
        .where((r) => r.device.remoteId.str == targetDevice)
        .firstOrNull
        ?.device;
    if (device != null) {
      await device?.connect();
      'Device named ${device?.platformName} has been connected'.log;
      setDevice?.call(device!);
      return true;
    }

    if (!isStrict && device == null) {
      device = scanned.firstOrNull?.device;
    }

    if (device != null) {
      await device?.connect();
      'Device named ${device?.platformName} has been connected'.log;
      setDevice?.call(device!);
      return true;
    }

    /// 3. fail to connect device
    return false;
  }

  Future<void> disconnectDevice({BluetoothDevice? device}) async {
    /// simulator
    if (isSimulator) {
      isSimulatorConnected = false;
      return;
    }

    if (device != null) {
      await device.disconnect();
      'Device named ${device.platformName} has been disconnected now'.log;
      this.device = null;
      return;
    }

    final connected = await FlutterBluePlus.systemDevices;
    'connected:\n$connected'.log;

    await Future.wait(
      [
        for (final device in connected.where(scanFilter))
          () async {
            await device.disconnect();
            'Device named ${device.platformName} has been disconnected now'.log;
          }(),
      ],
    );
    this.device = null;
  }

  void _setupBleStreams() {
    streams.ble
      ..connection = _getConnectionStateStream().broadcast
      ..battery = _getBatteryLevelStream().broadcast
      ..tap = _getPhysicalTapStream().broadcast;
  }

  Future<void> registerCharacteristics() async {
    /// simulator
    if (isSimulator) return _setupBleStreams();

    /// Non-simulator
    if (device == null) return;

    final services = await device!.discoverServices();
    for (final service in services) {
      final characteristics = service.characteristics;
      for (final characteristic in characteristics) {
        uuid.characteristic.forEach(
          (key, value) {
            final uuid = characteristic.uuid.str128;
            if (value == null) return;
            if (uuid.toLowerCase() == value.toLowerCase()) {
              '${key.name} was registered [${uuid.toLowerCase()}]'.log;
              points[key] = characteristic;
            }
          },
        );
      }
    }
    _setupBleStreams();
  }

  Stream<bool> _getConnectionStateStream() async* {
    /// simulator
    if (isSimulator) {
      yield* Stream.periodic(
        const Duration(milliseconds: 500),
        (_) => isSimulatorConnected,
      );
      return;
    }

    /// non-simulator
    final stream = device?.connectionState.broadcast;
    if (stream == null) return;

    var prev = BluetoothConnectionState.disconnected;
    final eventStream = Rx.merge(
      [stream, Stream.periodic(const Duration(milliseconds: 500), (_) => prev)],
    );

    await for (final event in eventStream) {
      yield event == BluetoothConnectionState.connected;
      prev = event;
    }

    yield false;
  }

  Stream<int?> _getBatteryLevelStream() async* {
    /// simulator
    if (isSimulator) {
      yield* Stream.periodic(
        const Duration(seconds: 1),
        (_) => isSimulatorConnected ? 100 : null,
      );
      return;
    }

    /// non-simulator
    final battery = points[Characteristics.BATTERY_LEVEL];
    if (battery == null) return;

    try {
      await battery.setNotifyValue(true);
    } catch (e) {
      '_getBatteryLevelStream: $e'.log;
      try {
        await Future<void>.delayed(const Duration(seconds: 2));
        await battery.setNotifyValue(true);
      } catch (e) {
        '_getBatteryLevelStream (second try): $e'.log;
      }
    }

    var prev = await battery.read();
    final stream = battery.lastValueStream.broadcast;
    final eventStream = Rx.merge(
      [stream, Stream.periodic(const Duration(seconds: 1), (_) => prev)],
    );
    await for (final event in eventStream) {
      if (device == null) break;
      if (event.isEmpty) continue;
      yield event[0];
      prev = event;
    }
    yield null;
  }

  Stream<TapEvent?> _getPhysicalTapStream() async* {
    /// simulator
    if (isSimulator) return;

    final tap = points[Characteristics.TAP_BUTTON];
    if (tap == null) return;

    try {
      await tap.setNotifyValue(true);
    } catch (e) {
      '_getPhysicalTapStream: $e'.log;
      try {
        await Future<void>.delayed(const Duration(seconds: 2));
        await tap.setNotifyValue(true);
      } catch (e) {
        '_getPhysicalTapStream (second try): $e'.log;
      }
    }

    TapEvent? prev;
    final stream = tap.onValueReceived.broadcast;
    final eventStream = Rx.merge(
        [stream, Stream.periodic(const Duration(seconds: 1), (_) => prev)]);

    await for (final event in eventStream) {
      if (device == null) break;
      if (prev == event) continue;
      prev = TapEvent(
        deviceName: device?.platformName,
        time: DateTime.now(),
      );
      yield prev;
    }
    yield null;
  }
}
