import 'package:copy_with_extension/copy_with_extension.dart';
import 'package:healthrian_ble_interface/healthrian_ble_interface.dart';
import 'package:healthrian_ble_interface/src/extension/extension.dart';
import 'package:lzma/lzma.dart';
import 'package:rxdart/rxdart.dart';

part 'module_ecg.g.dart';

typedef CallbackType = (
  void Function(Channels, DataFilter)? callback,
  int period
);

@CopyWith()
class ModuleEcg {
  ModuleEcg({
    required this.ble,
    this.commands = const Commands(),
    this.udfRawStream = Builtin.udfRawStream,
    this.udfConvert = Builtin.udfConvertWearEcg,
    this.udfLeadStream = Builtin.udfLeadStreamWearEcg,
    this.udfLeadCompute = Builtin.udfLeadComputeWearEcg,
    this.udfFilter = Builtin.udfFilter,
    this.channelCount = 12,
    this.sampleCount = 9,
    this.filterCount = 8,
  });

  final ModuleBle ble;
  final Commands commands;

  final Stream<List<int>> Function({
    required ModuleEcg ecg,
  }) udfRawStream;

  final List<List<double>>? Function({
    required List<int> event,
    required int channelCount,
    required int sampleCount,
  }) udfConvert;

  final Stream<List<int>> Function({
    required ModuleEcg ecg,
  }) udfLeadStream;

  final LeadState? Function({
    required List<int> event,
  }) udfLeadCompute;

  final Channels Function({
    required Channels channels,
    required int channelCount,
    required int sampleCount,
    required DataFilter dataFilter,
    required bool isAttached,
  }) udfFilter;

  bool isFilterAlways = true;

  /// [samplingRate] is the number samples per second
  int get samplingRate =>
      dataFilter.samplingRate == SamplingRate.freq250hz ? 250 : 500;

  /// [channelCount] is the number of channels
  final int channelCount;

  /// [sampleCount] is the number of samples in a packet
  final int sampleCount;

  /// [filterCount] is the number of data filters
  final int filterCount;

  bool isStrictCallback = false;
  bool isCleanerCallback = false;

  /// [maxBuffer] decides the maximum length of buffered Stream.
  double maxBuffer = 40;

  /// [lengthBuffer] decides a length of buffered Stream,
  double lengthBuffer = 30;

  /// [displayBuffer] decides a length of buffered Stream,
  /// which is shown to screen.
  double displayBuffer = 5;

  /// [refreshBuffer] decides a length to be hidden after refreshed.
  double refreshBuffer = 0.5;

  /// If [isCompressed] is `true`,
  bool isCompressed = false;

  DataFilter dataFilter = DataFilter(8);

  AdaptiveHeartRate adaptiveHR = AdaptiveHeartRate();

  final List<CallbackType> callbacks = [];

  /// [lastChannel] is a latest stream value from [ble.streams.bufferedEcg]
  Channels? lastChannel;

  /// [lastLeadState] is a latest stream value from [ble.streams.leadStateEcg]
  LeadState? lastLeadState;

  bool isPaused = false;
  bool isMeasuring = false;

  ///////////////////////////////////////////////////////////////////////

  void initStream() {
    ble.streams.ecg.preload
      ..raw = null
      ..lzma = null
      ..event = null
      ..value = null
      ..leadState = null
      ..heartRate = null;
    ble.streams.ecg.original
      ..buffer = null
      ..plot = null
      ..plotRT = null;
    ble.streams.ecg.filter
      ..buffer = null
      ..plot = null
      ..plotRT = null;
  }

  Future<void> start() async {
    dataFilter = DataFilter(filterCount);
    adaptiveHR = AdaptiveHeartRate(is250Hz: samplingRate == 250);

    await _sendDate();
    await _sendStartSave();
    await _sendStart();

    isMeasuring = true;
    isPaused = false;

    ble.streams.ecg.preload
      ..raw = _getRaw().broadcast
      ..lzma = _getLzma().broadcast
      ..event = _getEvent()?.broadcast
      ..value = _getConverted().broadcast
      ..leadState = _getLeadState().broadcast
      ..heartRate = _getHeartRate().broadcast;
    ble.streams.ecg.original
      ..buffer = _getBuffered().broadcast
      ..plot = _getPlot(isFiltered: false).broadcast
      ..plotRT = _getPlotRT(isFiltered: false).broadcast;
    ble.streams.ecg.filter
      ..buffer = _getBufferedFilter().broadcast
      ..plot = _getPlot(isFiltered: true).broadcast
      ..plotRT = _getPlotRT(isFiltered: true).broadcast;
  }

  void resetPlot() {
    ble.streams.ecg.original
      ..buffer = _getBuffered().broadcast
      ..plot = _getPlot(isFiltered: false).broadcast
      ..plotRT = _getPlotRT(isFiltered: false).broadcast;
    ble.streams.ecg.filter
      ..buffer = _getBufferedFilter().broadcast
      ..plot = _getPlot(isFiltered: true).broadcast
      ..plotRT = _getPlotRT(isFiltered: true).broadcast;
  }

  /// [pause]
  Future<void> pause() async {
    isPaused = !isPaused;
  }

  /// [stop] sends stop_ecg_packet to device.
  Future<void> stop() async {
    isMeasuring = false;
    final control = ble.points[Characteristics.STOP_ECG];
    await control.writePacket(commands.stopEcg);
  }

  Future<void> _sendStart() async {
    final control = ble.points[Characteristics.START_ECG];
    await control.writePacket(commands.startEcg);
  }

  Future<void> _sendDate() async {
    final control = ble.points[Characteristics.SET_DATA];
    await control.writePacket(commands.setDate);
  }

  Future<void> _sendStartSave() async {
    final control = ble.points[Characteristics.CONTROL_POINT];
    await control.writePacket(commands.save.startEcg);
  }

  Stream<List<int>> _getRaw() async* {
    final events = Rx.merge(
      [
        udfRawStream(ecg: this),
        Stream.periodic(const Duration(seconds: 1), (_) => const <int>[])
      ],
    );

    await for (final event in events) {
      if (!isMeasuring) break;
      if (!ble.isSimulator && ble.device == null) break;
      if (ble.isSimulator && !ble.isSimulatorConnected) break;
      if (isPaused) continue;
      if (event.isEmpty) continue;
      yield event;
    }
  }

  Stream<List<int>> _getLzma() async* {
    final events = ble.streams.ecg.preload.raw;
    if (events == null) return;
    await for (final event in events) {
      yield lzma.decode(event);
    }
  }

  Stream<List<int>>? _getEvent() async* {
    final events = isCompressed
        ? ble.streams.ecg.preload.lzma
        : ble.streams.ecg.preload.raw;
    if (events == null) return;
    await for (final event in events) {
      yield event;
    }
  }

  /// list double
  Stream<List<List<double>>> _getConverted() async* {
    final events = ble.streams.ecg.preload.event;
    if (events == null) return;
    await for (final event in events) {
      final value = udfConvert(
          event: event, channelCount: channelCount, sampleCount: sampleCount);
      if (value == null) continue;
      yield value;
    }
  }

  ///
  Stream<LeadState?> _getLeadState() async* {
    List<int>? prev;
    const sec = Duration(milliseconds: 1000);
    const half = Duration(milliseconds: 500);
    final events = Rx.merge(
      [
        udfLeadStream(ecg: this),
        Stream.periodic(sec, (_) => prev ?? <int>[]),
      ],
    ).throttleTime(half);

    await for (final event in events) {
      if (!isMeasuring) break;
      if (!ble.isSimulator && ble.device == null) break;
      if (ble.isSimulator && !ble.isSimulatorConnected) break;
      if (isPaused) continue;
      if (event.isEmpty) continue;

      prev = event;

      final leadState = udfLeadCompute(event: event);
      if (leadState == null) continue;
      yield leadState;

      lastLeadState = leadState;
    }
    yield null;
  }

  ///
  Stream<int?> _getHeartRate() async* {
    final events = ble.streams.ecg.preload.value;
    if (events == null) return;
    var last = DateTime.now();
    await for (final event in events) {
      final l2 = event[event.length >= 2 ? 1 : 0];
      for (var i = 0; i < l2.length; i++) {
        final isAttachedNormal = lastLeadState?.isAttached ?? true;
        final isAttached = isFilterAlways || isAttachedNormal;

        final output = adaptiveHR.calculateHR(l2[i], isAttached);

        final hr = output <= 0 ? null : output;
        final diff = DateTime.now().difference(last);
        if (hr == null && diff < const Duration(seconds: 2)) continue;
        yield hr;
        last = DateTime.now();
      }
    }
    yield null;
  }

  Stream<Channels> _getBuffered() async* {
    final events = ble.streams.ecg.preload.value;
    if (events == null) return;

    final lengthBufferSize = (lengthBuffer * samplingRate).round();
    final maxBufferSize = (maxBuffer * samplingRate).round();

    final res = List.generate(channelCount, (index) => <double>[]);

    var x = 0.0;

    final callbacksTargets = <int>[];

    await for (final event in events) {
      for (var k = 0; k < channelCount; k++) {
        res[k].addAll(event[k]);
      }

      for (final channel in res) {
        if (channel.length > maxBufferSize) {
          channel.removeRange(0, channel.length - lengthBufferSize);
        }
      }

      final channels = Channels.fromList(res);
      yield channels;

      if (callbacksTargets.isEmpty) {
        callbacksTargets.addAll(
          List.generate(callbacks.length, (index) => callbacks[index].$2),
        );
      }

      for (var i = 0; i < callbacks.length; i++) {
        final callback = callbacks[i].$1;
        final period = callbacks[i].$2;
        if (callback != null && x >= callbacksTargets[i]) {
          if (isStrictCallback) {
            final start = channels.l2.length - period - (x.round() % period);
            final end = channels.l2.length - (x.round() % period);
            callback.call(channels.range(start, end), dataFilter);
          } else {
            callback.call(channels, dataFilter);
          }
          callbacksTargets[i] += period;
        }
      }

      lastChannel = channels;
      x += sampleCount;
    }

    if (!isCleanerCallback) return;
    for (var i = 0; i < callbacks.length; i++) {
      final callback = callbacks[i].$1;
      final period = callbacks[i].$2;
      if (callback != null && x < callbacksTargets[i]) {
        if (isStrictCallback) {
          final start = lastChannel!.l2.length - (x.round() % period);
          final end = lastChannel!.l2.length;
          callback.call(lastChannel!.range(start, end), dataFilter);
        } else {
          callback.call(lastChannel!, dataFilter);
        }
        callbacksTargets[i] += period;
      }
    }
  }

  Stream<Channels> _getBufferedFilter() async* {
    final events = ble.streams.ecg.original.buffer;
    if (events == null) return;

    final maxBufferSize = (maxBuffer * samplingRate).round();
    final lengthBufferSize = (lengthBuffer * samplingRate).round();

    final res = List.generate(channelCount, (index) => <double>[]);

    await for (final event in events) {
      final isAttachedNormal = lastLeadState?.isAttached ?? true;
      final isAttached = isFilterAlways || isAttachedNormal;

      final startIndex = event.l2.length - sampleCount;
      final endIndex = event.l2.length;
      for (var k = 0; k < channelCount; k++) {
        res[k].addAll(event.list[k].getRange(startIndex, endIndex));
      }
      for (final channel in res) {
        if (channel.length > maxBufferSize) {
          channel.removeRange(0, channel.length - lengthBufferSize);
        }
      }
      final channels = Channels.fromList(res);

      yield udfFilter(
        channels: channels,
        channelCount: channelCount,
        sampleCount: sampleCount,
        dataFilter: dataFilter,
        isAttached: isAttached,
      );
    }
  }

  /// OOO OOO OOO OOO NNN
  /// XXX OOO OOO OOO OOO
  ///     OOO OOO OOO OOO
  Stream<PlotChannels> _getPlot({required bool isFiltered}) async* {
    final events = isFiltered
        ? ble.streams.ecg.filter.buffer
        : ble.streams.ecg.original.buffer;

    if (events == null) return;

    final refreshBufferSize = (refreshBuffer * samplingRate).round();

    final resList = List.generate(12, (index) => <FlSpot>[], growable: false);
    final res =
        PlotChannels.fromList(resList, 0, refreshBufferSize, sampleCount);

    var x = 0.0;
    await for (final event in events) {
      for (var i = event.l2.length - sampleCount; i < event.l2.length; i++) {
        final j = i - event.l2.length + sampleCount;

        for (var k = 0; k < channelCount; k++) {
          res.list[k].add(FlSpot(x + j, event.list[k][i]));
        }
      }

      if (res.l2.length > event.l2.length) {
        for (var k = 0; k < channelCount; k++) {
          res.list[k].removeRange(0, res.list[k].length - event.list[k].length);
        }
      }

      yield res;
      x += sampleCount;
    }
  }

  ///     NNN
  /// OOO XXX OOO OOO
  /// OOO OOO OOO OOO
  Stream<PlotChannels?> _getPlotRT({required bool isFiltered}) async* {
    final events = isFiltered
        ? ble.streams.ecg.filter.plot
        : ble.streams.ecg.original.plot;

    if (events == null) return;

    final refreshBufferSize = (refreshBuffer * samplingRate).round();
    final displayBufferSize = (displayBuffer * samplingRate).round();

    double skip(double x) => x % displayBufferSize;
    FlSpot f(FlSpot original) => FlSpot(skip(original.x), original.y);

    final resList = List.generate(
      12,
      (_) => List<FlSpot>.generate(
        displayBufferSize,
        (index) => FlSpot.nullSpot,
        growable: false,
      ),
    );

    final res =
        PlotChannels.fromList(resList, 0, refreshBufferSize, sampleCount);
    var x = 0;

    await for (final event in events) {
      final start = event.l2.length - sampleCount;
      final end = event.l2.length;
      for (var i = start; i < end; i++) {
        final j = i - event.l2.length + sampleCount;
        final index = (x + j.toDouble()).round() % displayBufferSize;

        for (var k = 0; k < channelCount; k++) {
          res.list[k][index] = f(event.list[k][i]);
        }
      }
      x += sampleCount;
      yield res.copyWith(pivot: (x % displayBufferSize));
    }
    yield null;
  }
}
